#include <FreeRTOS.h>
#include <task.h>
#include <portable.h>
#include <queue.h>
#include <string.h>

#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/exti.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/cm3/nvic.h>

#include <lib/mini-printf.h>

#include "gpio_utils.h"
#include "serial_utils.h"

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

//add stack-overflow hook
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName);

// our tasks
void task_onboardled(void *args);
void task_buttonled(void *args);
void task_uart_tx(void *args);
void task_uart_rx(void *args);
static QueueHandle_t btn_q; // coda dall'ISR del pulsante a task_buttonled
static QueueHandle_t uart_tx_q;
void dds_send_byte(unsigned char byte);
void dds_send_word(uint16_t byte);
void dds_test(void);

/*
 * Partiamo dall'esempio `serialpc` e aggiungiamo il controllo del DDS sulle porte PB_15 e PB_13. Queste porte
 * sono anche utilizzabili come SPI2 (rispettivamente MOSI e SCK), ma in questo esempio facciamo bit-banging.
 * Scegliamo però le porte "giuste" in modo che quando faremo la versione SPI abbiamo tutto fatto giusto.
 */

static void gpio_setup(void)
{

	// built-in LED
	rcc_periph_clock_enable(RCC_GPIOC);
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO13);

	rcc_periph_clock_enable(RCC_GPIOA);
	rcc_periph_clock_enable(RCC_GPIOB);

    // Serial comunication
    rcc_periph_clock_enable(RCC_USART1);
    gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ,
            GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
            GPIO_USART1_TX);
    usart_set_baudrate(USART1, 38400);
    usart_set_databits(USART1,8);
    usart_set_stopbits(USART1,USART_STOPBITS_1);
    usart_set_mode(USART1,USART_MODE_TX_RX);
    usart_set_parity(USART1,USART_PARITY_NONE);
    usart_set_flow_control(USART1,USART_FLOWCONTROL_NONE);
    usart_enable(USART1);

    rcc_periph_clock_enable(RCC_SPI1);
	gpio_set_mode(GPIOB, GPIO_MODE_OUTPUT_50_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO15 | GPIO13 | GPIO12);
	gpio_set_mode(GPIOB, GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO14);
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO11);

    /*
    spi_reset(SPI2);
    spi_init_master(
            SPI2,
            SPI_CR1_BAUDRATE_FPCLK_DIV_256,
            SPI_CR1_CPOL_CLK_TO_0_WHEN_IDLE,
            SPI_CR1_CPHA_CLK_TRANSITION_1,
            SPI_CR1_DFF_8BIT,
            SPI_CR1_LSBFIRST
            );
    spi_set_standard_mode(SPI2, 2);
    //spi_set_bidirectional_transmit_only_mode(SPI2);
    spi_enable(SPI2);
    spi_disable_software_slave_management(SPI2);
    spi_enable_ss_output(SPI2);
    */
}

void dds_send_word(uint16_t byte) {
    uint16_t origbyte = byte;
    char *buf = (void *)pvPortMalloc(24);
    for(int i=0; i<16; i++) { // every bit
        char bit = (byte & 0x8000) != 0;
        if(buf) buf[i] = bit ? '1': '0';
        if(bit) {
            gpio_set(GPIOB, GPIO15);
        } else {
            gpio_clear(GPIOB, GPIO15);
        }
        gpio_pulseLow(GPIOB, GPIO13);
        byte = byte << 1;
    }
    if (buf) {
        snprintf(buf+16, 8, "\t%x\n", origbyte);
        if (xQueueSend(uart_tx_q, (void *)&buf, 0) != pdTRUE) {
            vPortFree(buf);
        }
    }
}

void dds_test(void) {
    gpio_clear(GPIOA, GPIO11);
    for(int i=0; i<100; i++) __asm__("nop");
       dds_send_word(0x2100);
       dds_send_word(0x51b7);
       dds_send_word(0x4003);
       dds_send_word(0xC000);
       dds_send_word(0x2000);
       /*
    //long freq = 1000;
    //long freqWord = (freq << 28) / (25000000);
    //uint16_t msb = ((freqWord & 0xFFFC000) >> 14);
    //uint16_t lsb = (freqWord & 0x3FFF);
    spi_write(SPI2, 0x2100);
    spi_write(SPI2, 0x29f1);
    spi_write(SPI2, 0); //phase
    spi_write(SPI2, 0xC000); //waveform
    spi_write(SPI2, 0x2000);
    */
    for(int i=0; i<10; i++) __asm__("nop");
    gpio_set(GPIOA, GPIO11);
}

static void uart_puts(const char *s) {
    for(; *s; ++s) {
        usart_send_blocking(USART1, *s);
    }
}


int main(void)
{
    rcc_clock_setup_in_hse_8mhz_out_72mhz();
	gpio_setup();
    gpio_set(GPIOA, GPIO8);
    gpio_clear(GPIOB, GPIO11);
    uart_puts("READY\n");

    btn_q = xQueueCreate(10, sizeof(char));
    uart_tx_q = xQueueCreate(20, sizeof(char*));
	xTaskCreate(task_buttonled, "BUTTONLED", 300, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_onboardled, "ONBOARD", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_uart_tx, "TX", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	xTaskCreate(task_uart_rx, "RX", 50, NULL, configMAX_PRIORITIES - 1,
	            NULL);
	vTaskStartScheduler();
	for (;;) {
	}

	return 0;
}

int boardLedDuty = 1000;
void task_onboardled(void *UNUSED(args))
{
	while (1) {
        dds_test();
        q_printf(uart_tx_q, 16, "ci provo\n");
        gpio_clear(GPIOC, GPIO13); /* LED on */
        vTaskDelay(pdMS_TO_TICKS(boardLedDuty));

		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(500));
	}
}

void task_buttonled(void *UNUSED(args))
{
    while (1) {
        char increasing;
        if(xQueueReceive(btn_q, (void*)&increasing, pdMS_TO_TICKS(5000)) == pdPASS)
        {
            if(increasing) {
                gpio_toggle(GPIOA, GPIO8);
            } else {
                gpio_toggle(GPIOB, GPIO11);
            }
            char *buf = (void*) pvPortMalloc(16);
            if(buf != NULL) {
                snprintf(buf, 16, "%c %d\n", increasing ? 'U' : 'D', boardLedDuty);
                // attenzione: xQueueSend vuole un puntatore alle cose che deve copiare.
                // noi dobbiamo mandare una stringa, quindi un puntatore. Dobbiamo quindi mandare un puntatore
                // a puntatore!
                if(xQueueSend(uart_tx_q, (void*) &buf, 0) != pdTRUE) {
                    // la free la fa il task che legge; ma se non riusciamo a mettere in coda (coda piena) 
                    // allora la free dobbiamo farla noi
                    vPortFree(buf);
                }
            }
        }
    }
}

void task_uart_tx(void *UNUSED(args)) {
    while (1) {
        char *buf, *buf2;
        if(xQueueReceive(uart_tx_q, (void*)&buf, pdMS_TO_TICKS(5000)) != pdPASS)
            continue;
        buf2 = buf; // mi devo tenere il puntatore originale senno' dopo non posso fare free
        for(; *buf != '\0'; buf++) {
            while(!usart_get_flag(USART1, USART_SR_TXE))
                taskYIELD();
            usart_send_blocking(USART1, *buf);
        }
        vPortFree(buf2);
    }
}

void task_uart_rx(void *UNUSED(args)) {
    char cmd[32];
    unsigned char i = 0;
    while (1) {
        if (i >= 30) { // avoid overflows
            i = 0;
        }
        cmd[i] = usart_recv_blocking(USART1);
        if (cmd[i] == '\r' || cmd[i] == '\n') {
            gpio_toggle(GPIOB, GPIO11);
            cmd[i] = '\0'; // replace newline with end-of-string
            if (strcmp(cmd, "GET") == 0) { //check valid commands
                char *buf = (void *)pvPortMalloc(16);
                if (buf != NULL) {
                    snprintf(buf, 16, "-> %d\n", boardLedDuty);
                    if (xQueueSend(uart_tx_q, (void *)&buf, 0) != pdTRUE) {
                        vPortFree(buf);
                    }
                }
            } // GET
            i=0;
        } // newline
        else {
            i++;
        }
    }
}

// three "small flashes" (an S in Morse code) on onboard led
//
// questa funzione viene chiamata in caso di stackoverflow su uno dei task; e'
// quindi utile metterci un "segnale" riconoscibile per debug. In questo caso 3
// flash rapidi.
//
// siccome "tarare" la dimensione dello stack necessario e' difficile, con
// questa funzione si puo' andare "a tentoni" e ridurre la dimensione finche'
// non si ottengono crash
//
// in questa funzione, a quanto pare (ma non ho trovato doc) non si possono
// usare cose come vTaskDelay. Siamo quindi privi di RTOS e per fare gli sleep
// dobbiamo fare i for "all'antica"

void vApplicationStackOverflowHook(xTaskHandle *UNUSED(pxTask),
        signed portCHAR *UNUSED(pcTaskName))
{
    int i;
    for (;;) {
        for (char c = 0; c < 3; c++) {
            gpio_clear(GPIOC, GPIO13); /* LED on */
            for (i = 0; i < 1000000; i++)   /* small delay */
                __asm__("nop");

            gpio_set(GPIOC, GPIO13);   /* LED off */
            for (i = 0; i < 3000000; i++) { /* a bit more delay */
                __asm__("nop");
            }
        }
        for (i = 0; i < 10000000; i++) {    /* long delay */
            __asm__("nop");
        }
    }
}

// vim: set fdm=marker:
