/* Demo program for OLED 128x64 SSD1306 controller
 * Warren Gay   Sun Dec 17 22:49:14 2017
 *
 * Important!  	You must have a pullup resistor on the NSS
 * 	       	line in order that the NSS (/CS) SPI output
 *		functions correctly as a chip select. The
 *		SPI peripheral configures NSS pin as an
 *		open drain output.
 *
 * OLED		4-Wire SPI
 *
 * PINS:
 *	PC13	LED
 *	PA15	/CS (NSS, with 10k pullup)
 *	PB3	SCK
 *	PB5	MOSI (MISO not used)
 *	PB10	D/C
 *	PB11	/Reset
 */
#include <string.h>
#include <ctype.h>

#include "mcuio.h"
#include "miniprintf.h"
#include "intelhex.h"

#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
//#include "meter.h"
#include "oled.h"

#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/gpio.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/stm32/spi.h>

#define SSD_SPI  SPI2
#define SSD_SCK  GPIO13
#define SSD_MOSI GPIO15
#define SSD_DC   GPIO10
#define SSD_RST  GPIO11
#define SSD_PORT  GPIOB
// il CS sta su porta GPIOA! bisognerebbe mettere tutto su porta GPIOB per semplicità
#define SSD_CS   GPIO12

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

void setup(void);
void task_onboardled(void *UNUSED(args));
	


/*
 * Monitor task:
 */
static void
monitor_task(void *arg __attribute((unused))) {
	float v = 1.3;
	
	std_puts("\nMonitor Task Started.\n");

    struct ssd1306 oled = {
        SSD_SPI, GPIOB, SSD_SCK, SSD_MOSI, SSD_CS, SSD_DC, SSD_RST
    };
	oled_init(&oled);
    std_puts("oled  OK\n");
    std_puts("meter OK\n");


    uint8_t graph[128*64/8];
    for(int i=0; i<128*64/8; i++) {
        graph[i] = 17;
    }
    for (;;) {
		vTaskDelay(pdMS_TO_TICKS(600));
        v += 0.1;
        if (v > 3) {
            v = 0.1;
        }
        //meter_set_value(&m1, v);
        //meter_update();
        oled_draw(&oled, graph);
    }
}

void setup(void) {
	rcc_periph_clock_enable(RCC_GPIOA);
	rcc_periph_clock_enable(RCC_GPIOB);
	rcc_periph_clock_enable(RCC_GPIOC);
	rcc_periph_clock_enable(RCC_SPI2);

	// LED on PC13
	gpio_set_mode(GPIOC, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO13);
	gpio_set(GPIOC,GPIO13);	// PC13 = on

    // usart {{{
    gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ, GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
             GPIO_USART2_TX);
    open_uart(2,38400,"8N1","rw",0,0);
    // usart }}}


    // SSD1306 {{{
	// PB10 -> D/C, PB11 -> RES
	gpio_set_mode(GPIOB, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL,
		SSD_DC|SSD_RST);
	// Activate OLED Reset line
	gpio_clear(GPIOB,SSD_RST);

    gpio_set_mode( GPIOB, GPIO_MODE_OUTPUT_50_MHZ,
            GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
            SSD_SCK|SSD_MOSI
            );
	// NSS=PA15
	gpio_set_mode( SSD_PORT, GPIO_MODE_OUTPUT_50_MHZ,
        	GPIO_CNF_OUTPUT_ALTFN_PUSHPULL, SSD_CS);
    // SSD1306 }}}
}

int main(void) {

	rcc_clock_setup_in_hse_8mhz_out_72mhz();	// Blue pill

    setup();
	std_set_device(mcu_uart2);


	spi_reset(SSD_SPI);
	spi_init_master(
		SSD_SPI,
                SPI_CR1_BAUDRATE_FPCLK_DIV_256,
                SPI_CR1_CPOL_CLK_TO_0_WHEN_IDLE,
		SPI_CR1_CPHA_CLK_TRANSITION_1,
	        SPI_CR1_DFF_8BIT,
	        SPI_CR1_MSBFIRST
	);
	spi_disable_software_slave_management(SSD_SPI);
	spi_enable_ss_output(SSD_SPI);

	usb_start(1,1);
	gpio_clear(GPIOC,GPIO13);			// PC13 = off

	xTaskCreate(task_onboardled, "board", 100, NULL, configMAX_PRIORITIES - 1, NULL);
	xTaskCreate(monitor_task,"monitor",1000,NULL, configMAX_PRIORITIES -1,NULL);
	vTaskStartScheduler();
	for (;;);
	return 0;
}

void task_onboardled(void *UNUSED(args))
{
	while (1) {
		gpio_clear(GPIOC, GPIO13); /* LED on */
		vTaskDelay(pdMS_TO_TICKS(600));
		gpio_set(GPIOC, GPIO13); /* LED off */
		vTaskDelay(pdMS_TO_TICKS(200));
	}
}


// End
